/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * generator.cc
 * Copyright (C) 2016 Fran??ois Gautrais <francois@gautrais.eu>
 *
 * histodex is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * histodex is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "generator.h"
#include "os.h"
#include "image-obj.h"
#include "g-window.h"
#include "resource-tab.h"
#include "stage-tab.h"
#include "zone-tab.h"
#include "general-tab.h"
#include "stage.h"
#include "resource.h"
#include "area.h"
#include "general.h"

#include <cstdlib>
#include "utils.h"

#define FILE_AUDIO_OBJ "models/audio/audio.obj"
#define FILE_AUDIO_MTL "models/audio/audio.mtl"
#define FILE_VIDEO_OBJ "models/camera/camera.obj"
#define FILE_VIDEO_MTL "models/camera/camera.mtl"
 
 
Generator::Generator(General& g, std::vector<Area>& a,std::vector<Resource>& r,
                     std::vector<Stage>& s, GWindow* ui)
{
	m_gen=g;
	m_resources=r;
	m_areas=a; 
	m_stages=s;
	m_ui=ui;
	m_nErrors=0;
	m_nWarning=0;
	m_nCriticalErrors=0;
	std::cout << "Warn =" << m_nWarning << "\n";
	std::cout << "Err =" << m_nErrors << "\n";
	prepare();
}

Generator::~Generator()
{
} 

std::string Generator::getLog()
{
	int e=0, c=0, w=0;
	char tmp[256];
	std::string ret = "Log:\n";
	for(int i=0; i<m_log.size(); i++)
		ret+=m_log[i]+"\n";
	ret+="\n\nErreurs\n";
	
	for(int i=0; i<m_errors.size(); i++)
	{ 
		if(m_errors[i][0] == 'A') w++;
		else if(m_errors[i][0] == 'E') e++;
		else c++;
		ret+=m_errors[i]+"\n";
	}
	sprintf(tmp, "\nTotal:\n\tWarnings: %d\n\tErreurs: %d\n\tErreurs critiques: %d\n",
	    		m_nWarning, m_nErrors, m_nCriticalErrors);
	ret+=tmp;
	return ret;
}

void Generator::log(const std::string& l)
{
	m_log.push_back (l);
}
 
void Generator::error(const std::string& l, int level)
{
	std::string p;
	if(level==LOG_WARNING) { p="Attention: "; m_nWarning++;}
	else if(level==LOG_ERROR) {p="Erreur: "; m_nErrors++;}
	else {p="Erreur Critique: "; m_nCriticalErrors++; }
	m_errors.push_back (p+l);
} 

void Generator::generateMap()
{ 
	std::string map = m_gen.getMap ();
	if(map!="")
	{
		log("Génération du la map: '"+map+"' vers '" +OS::concatPath(m_pathRaw, strtolower (m_gen.m_name+"_map"))+"'");
		OS::copy(map, OS::concatPath(m_pathRaw, "map") );
	}else
	{
		log("Pas de carte a ajouter");
	}
}

void Generator::generate()
{
	Json::Value o = m_ui->getJson (true);
	std::cout << "gen: "<< OS::concatPath(m_pathRaw, strtolower (m_gen.m_name)) << std::endl;
	std::ofstream rout(OS::concatPath(m_pathRaw, strtolower (m_gen.m_name)));
	generateMap();
	for(int i=0; i<m_resources.size(); i++)
	{
		
		log("\nResource '"+m_resources[i].getName ()+"' type: "+m_resources[i].getType()+"");
		generateResource (m_resources[i]);
	}
	rout << o;
}   

Area*  Generator::findArea(const std::string& name)
{
	for(int i=0; i<m_areas.size(); i++)
		if(m_areas[i].getName () == name)
			return &m_areas[i];
	return NULL;
}

Resource*  Generator::findResource(const std::string& name)
{
	for(int i=0; i<m_resources.size(); i++)
		if(m_resources[i].getName () == name)
			return &m_resources[i];
	return NULL;
} 
   
Stage*  Generator::findStage(int i)
{
	for(int i=0; i<m_stages.size(); i++)
		if(m_stages[i].getStage () == i)
			return &m_stages[i];
	return NULL;
}

void Generator::check()
{
	checkStages();
	checkAreas();
	checkResources();
}

void Generator::checkStages()
{
	if(m_gen.m_nStages != m_stages.size())
		error("Nombre d'étapes incohérent", LOG_ERROR);
	for(int i=0; i<m_stages.size(); i++)
	{
		char tmp[8];
		sprintf(tmp, "%d", i);
		std::string r =  m_stages[i].getResource();
		if(findResource (r)==NULL)
		{
			error("La ressource '"+r+"' (étape "+std::string(tmp)+") n'existe pas", LOG_ERROR);
		}
	}
}

void Generator::checkAreas()
{
	for(int i=0; i<m_areas.size(); i++)
	{
		std::vector<std::string> names = split(m_areas[i].m_resource, ';');
		if( m_areas[i].getName() == "" )
		{
			error("Une zone n'a pas de nom", LOG_ERROR);
		}
		for(int j=0; j<names.size(); j++)
		{
			if(!findResource (names[j]))
				error("La zone '"+m_areas[i].getName()+"' a une ressource non valide ('"+names[j]+"')", LOG_ERROR);
		}
		if(!names.size())error("La zone '"+m_areas[i].getName()+"' n'a pas de ressource valide", LOG_ERROR );
	}
}

void Generator::checkResources()
{
	for(int i=0; i<m_resources.size(); i++)
	{
		if( m_resources[i].getName() == "" )
		{
			error("Une ressource n'a pas de nom", LOG_ERROR);
		}
		if(m_resources[i].getStage()>m_gen.m_nStages || (m_resources[i].getStage()<0))
		   error("La ressource '"+m_resources[i].getName()+"' n'a pas détape valide", LOG_ERROR);
	}
}

void Generator::prepare()
{
	std::string dir = m_gen.m_dir;
	std::string dir2;
	check();
	log("Création des répertoires:");

	/*

	log("\t"+OS::concatPath(dir, "app/"));
	dir= OS::concatPath(dir, "app/");
	OS::mkdir(dir);

	
	log("\t"+OS::concatPath(dir, "src/"));
	dir= OS::concatPath(dir, "src/");
	OS::mkdir(dir);

	log("\t"+OS::concatPath(dir, "main/"));
	dir= OS::concatPath(dir, "main/");
	OS::mkdir(dir);

	
	log("\t"+OS::concatPath(dir, "res/"));
	dir2=dir= OS::concatPath(dir, "res/");
	OS::mkdir(dir);

	log("\t"+OS::concatPath(dir, "drawable/"));
	dir2=OS::concatPath(dir, "drawable/");
	OS::mkdir(dir2);

	log("\t"+OS::concatPath(dir, "raw/"));
	dir= OS::concatPath(dir, "raw/");
	OS::mkdir(dir);*/

	log("\t"+OS::concatPath(dir, "assets/"));
	dir= OS::concatPath(dir, "assets/");
	OS::mkdir(dir);
	

	m_pathRaw=dir;
	m_pathDrawable=dir;
}

void Generator::generateResource3D(Resource& r)
{
	OS::Path o(r.getFile());
	std::ofstream rout(OS::concatPath(m_pathRaw, strtolower (r.getName()))+"_res");
	copyObj(r.getFile(), m_pathRaw, r.getName());
	log("\tGénération du fichier ressource: '"+OS::concatPath(m_pathRaw, strtolower (r.getName()))+"_res'");
	rout << r.getJson(true);
}

void Generator::generateResourceImage(Resource& r)
{
	ImageObj o (r.getFile(), m_gen.m_dir, r.getName (), this);
	o.generate ();
	log("\tGénération du fichier ressource: '"+OS::concatPath(m_pathRaw, strtolower (r.getName()))+"_res'");
	std::ofstream rout(OS::concatPath(m_pathRaw, strtolower (r.getName()))+"_res");
	rout << r.getJson(true);
}


void Generator::generateResourceAudio(Resource& r)
{
	OS::Path o(r.getFile());
	log("\tCopie du fichier audio '"+r.getFile()+"' vers '"+OS::concatPath(m_pathRaw, r.getName()+"_audio")+"'");
	std::ofstream rout(OS::concatPath(m_pathRaw, strtolower (r.getName()))+"_res");
	OS::copy(r.getFile(), OS::concatPath(m_pathRaw, r.getName()+"_audio") );
	log("\tGénération du fichier ressource: '"+OS::concatPath(m_pathRaw, strtolower (r.getName()))+"_res'");
	copyObj(FILE_AUDIO_OBJ, m_pathRaw, r.getName());
	rout << r.getJson(true);
}
 
  
void Generator::generateResourceVideo(Resource& r)
{
	OS::Path o(r.getFile());
	log("\tCopie du fichier video '"+r.getFile()+"' vers '"+OS::concatPath(m_pathRaw, r.getName()+"_video")+"'");
	std::ofstream rout(OS::concatPath(m_pathRaw, strtolower (r.getName()))+"_res");
	log("\tGénération du fichier ressource: '"+OS::concatPath(m_pathRaw, strtolower (r.getName()))+"_res'");
	copyObj(FILE_VIDEO_OBJ, m_pathRaw, r.getName());
	OS::copy(r.getFile(), OS::concatPath(m_pathRaw, r.getName()+"_video") );
	rout << r.getJson(true);
}
  


void Generator::generateResource(Resource& r)
{
	int n = stringtotype (r.getType());
	switch(n)
	{
		case TYPE_3D:
			generateResource3D (r);
			break;
		case TYPE_IMAGE:
			generateResourceImage (r); 
			break;
		case TYPE_AUDIO:
			generateResourceAudio (r);
			break;
		case TYPE_VIDEO:
		default:
			generateResourceVideo (r);
			break;
	}
}
 
std::string Generator::newFilename(const std::string& file)
{
	OS::Path p(file);
	return strtolower(p.m_name)+"_"+strtolower(p.m_ext);
}

 
void Generator::copyObj(const std::string& obj, const std::string& dir, const std::string& name)
{
	std::string line;
	OS::Path o(obj);
	std::ifstream in(obj);
	std::ofstream out(OS::concatPath(dir,strtolower (name)+"_obj"));
	std::cout << "OBJ " << dir+"/"+strtolower (name)+"_obj" << std::endl;
	log("\tCopie du modele 3D '"+obj+"' vers '"+OS::concatPath(dir,strtolower (name)+"_obj'"));
	while(std::getline(in, line))
	{
		int i, start;
		std::string buf="";
		for(i=0; i<line.size(); i++)
			if(line[i]!='\t' && line[i]!=' ')
				break;
		buf=line.substr(i);
		buf=buf;
		if( (start=buf.find("mtllib") ) != std::string::npos )
		{
			std::string tmp="";
			std::string file="";
			
			tmp=buf.substr(start, 7);
			for(int j=start+7; j<buf.size() && buf[j] != '\n'
			     && buf[j] != '\r'  && buf[j] != 0; j++)
				file+=buf[j];
			OS::Path p(file);
			std::cout << o.m_dir+file << std::endl;
			copyMtl(o.m_dir+file, m_pathRaw);
			
			out << tmp << strtolower (file) << std::endl;
		}
		else
			out << buf << std::endl;
	}
}   
 

int   Generator::getErrorsCount() { return m_nErrors; }
int   Generator::getWarningCount() { return m_nWarning; }
int   Generator::getCriticalErrorsCount(){ return m_nCriticalErrors; }
 
void Generator::copyMtl(const std::string& mtl, const std::string& dir)
{
	std::string line;
	OS::Path o(mtl);
	std::ifstream in(mtl);
	std::ofstream out(OS::concatPath(dir,strtolower (o.m_name)+".mtl"));
	//std::cout << "Mlt " << dir+"/"+strtolower (o.m_name)<< std::endl;
	log("\tCopie du fichier MTL '"+mtl+"' vers '"+OS::concatPath(dir,strtolower (o.m_name)+"_mtl'"));
	while(std::getline(in, line)) 
	{ 
		int i, start;
		std::string buf="";
		for(i=0; i<line.size(); i++)
			if(line[i]!='\t' && line[i]!=' ')
				break;
		buf=line.substr(i);
		buf=buf;
		if( (start=buf.find("map_Kd") )!= std::string::npos )
		{ 
			std::string tmp="";
			std::string file="";
 
			//std::cout << "buf = "<< buf << std::endl;
			tmp=buf.substr(start, 7);
			for(int j=start+7; j<buf.size() && buf[j] != '\n'
			     && buf[j] != '\r'  && buf[j] != 0; j++)
				file+=buf[j];
			OS::Path p(file);
			copyImage(o.m_dir+file, m_pathDrawable);
			
			out << tmp << strtolower (file) << std::endl;
		}
		else
			out << buf  << std::endl;
	}
}    
  
 
void Generator::copyImage(const std::string& img, const std::string& dir)
{
	std::string o="";
	OS::Path p(img);
	o=strtolower (p.m_name+"."+p.m_ext);
	log("\tCopie de la texture '"+img+"' vers '"+dir+o+"'");
	OS::copy(img, dir+o);
}
