/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * map-viewer.cc
 * Copyright (C) 2016 Fran??ois Gautrais <francois@gautrais.eu>
 *
 * histodex is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * histodex is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "map-viewer.h"
#include "zone-tab.h"
#include <gtk/gtk.h>

static OsmGpsMapSource_t opt_map_provider = OSM_GPS_MAP_SOURCE_OPENSTREETMAP;
static gboolean opt_friendly_cache = FALSE;
static gboolean opt_no_cache = FALSE;
static gboolean opt_debug = FALSE;
static char *opt_cache_base_dir = NULL;
static gboolean opt_editable_tracks = FALSE;
static GOptionEntry entries[] =
{
  { "friendly-cache", 'f', 0, G_OPTION_ARG_NONE, &opt_friendly_cache, "Store maps using friendly cache style (source name)", NULL },
  { "no-cache", 'n', 0, G_OPTION_ARG_NONE, &opt_no_cache, "Disable cache", NULL },
  { "cache-basedir", 'b', 0, G_OPTION_ARG_FILENAME, &opt_cache_base_dir, "Cache basedir", NULL },
  { "debug", 'd', 0, G_OPTION_ARG_NONE, &opt_debug, "Enable debugging", NULL },
  { "map", 'm', 0, G_OPTION_ARG_INT, &opt_map_provider, "Map source", "N" },
  { "editable-tracks", 'e', 0, G_OPTION_ARG_NONE, &opt_editable_tracks, "Make the tracks editable", NULL },
  { NULL }
};

gboolean on_save(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	MapViewer* mv = (MapViewer*) user_data;
	if(mv->m_init) mv->m_zone->on_save();
    return FALSE;
}

gboolean
on_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
	MapViewer* mv = (MapViewer*) user_data;
    OsmGpsMapPoint coord;
    float lat, lon;
    OsmGpsMap *map = OSM_GPS_MAP(widget);
    OsmGpsMapTrack *othertrack = OSM_GPS_MAP_TRACK(mv->m_rightclicktrack);

    int left_button =   (event->button == 1) && (event->button == 0);
    int middle_button = (event->button == 2) || ((event->button == 1) && (event->state & GDK_SHIFT_MASK));
    int right_button =  (event->button == 3) || ((event->button == 1) && (event->state & GDK_CONTROL_MASK));

    osm_gps_map_convert_screen_to_geographic(map, event->x, event->y, &coord);
    osm_gps_map_point_get_degrees(&coord, &lat, &lon);

    if (event->type == GDK_2BUTTON_PRESS) {
        if (right_button) {
			mv->set_place (lat, lon);
        }else if(middle_button)
		{
		}else
		{
			mv->add_point_to_poly (lat, lon);
		}
		if(mv->m_init) mv->m_zone->on_save();
    }

    return FALSE;
}

MapViewer::MapViewer(GtkBox* b, GtkSpinButton * sb, ZoneTab* z)
 :m_area(Json::Value(Json::arrayValue)), m_place(Json::Value(Json::arrayValue)), m_init(false)
{
	m_zone=z;
	m_angle = sb;
	m_polytrack=NULL;
	m_poly=NULL;
	m_repo_uri = osm_gps_map_source_get_repo_uri(opt_map_provider);
	m_cachebasedir = osm_gps_map_get_default_cache_directory();
	m_cachebasedir = osm_gps_map_get_default_cache_directory();

    if (opt_cache_base_dir && g_file_test(opt_cache_base_dir, G_FILE_TEST_IS_DIR)) {
        m_cachedir = g_strdup(OSM_GPS_MAP_CACHE_AUTO);
        m_cachebasedir = g_strdup(opt_cache_base_dir);
    } else if (opt_friendly_cache) {
        m_cachedir = g_strdup(OSM_GPS_MAP_CACHE_FRIENDLY);
    } else if (opt_no_cache) {
        m_cachedir = g_strdup(OSM_GPS_MAP_CACHE_DISABLED);
    } else {
        m_cachedir = g_strdup(OSM_GPS_MAP_CACHE_AUTO);
    }
	m_map = (OsmGpsMap*)g_object_new (OSM_TYPE_GPS_MAP,
                        "map-source",opt_map_provider,
                        "tile-cache",m_cachedir,
                        "tile-cache-base", m_cachebasedir,
                        "proxy-uri",g_getenv("http_proxy"),
                        NULL);

    m_osd = (OsmGpsMapLayer*)g_object_new (OSM_TYPE_GPS_MAP_OSD,
                        "show-scale",TRUE,
                        "show-coordinates",TRUE,
                        "show-crosshair",TRUE,
                        "show-dpad",TRUE,
                        "show-zoom",TRUE,
                        "show-gps-in-dpad",TRUE,
                        "show-gps-in-zoom",FALSE,
                        "dpad-radius", 30,
                        NULL);
    osm_gps_map_layer_add(OSM_GPS_MAP(m_map), m_osd);
    g_object_unref(G_OBJECT(m_osd));
	m_rightclicktrack = osm_gps_map_track_new();

	new_poly ();
    if(opt_editable_tracks)
        g_object_set(m_rightclicktrack, "editable", TRUE, NULL);
    osm_gps_map_track_add(OSM_GPS_MAP(m_map), m_rightclicktrack);

    g_free(m_cachedir);
    g_free(m_cachebasedir);

	osm_gps_map_set_keyboard_shortcut(m_map, OSM_GPS_MAP_KEY_FULLSCREEN, GDK_KEY_F11);
    osm_gps_map_set_keyboard_shortcut(m_map, OSM_GPS_MAP_KEY_UP, GDK_KEY_Up);
    osm_gps_map_set_keyboard_shortcut(m_map, OSM_GPS_MAP_KEY_DOWN, GDK_KEY_Down);
    osm_gps_map_set_keyboard_shortcut(m_map, OSM_GPS_MAP_KEY_LEFT, GDK_KEY_Left);
    osm_gps_map_set_keyboard_shortcut(m_map, OSM_GPS_MAP_KEY_RIGHT, GDK_KEY_Right);

	g_signal_connect (G_OBJECT (m_map), "button-press-event",
                G_CALLBACK (on_button_press_event), (gpointer) this);
	 
	g_signal_connect (G_OBJECT (m_map), "button-release-event",
                G_CALLBACK (on_save), (gpointer) this);
	 
	gtk_box_pack_end (GTK_BOX(b), GTK_WIDGET(m_map), TRUE, TRUE, 0);
	 m_init=true;
}

MapViewer::~MapViewer()
{
}




void MapViewer::centerView()
{
	if(m_place.size()>0)
	{
		osm_gps_map_set_center_and_zoom (m_map, (float)m_place[1].asDouble(),
		                             			(float)m_place[0].asDouble(),
		                             			16);
	}
	else if(m_area.size()>0)
	{
		
		osm_gps_map_set_center_and_zoom (m_map, (float)m_area[0][1].asDouble(),
		                             			(float)m_area[0][0].asDouble(),
		                             			16);
	}
}

void MapViewer::add_point_to_poly(float lat, float lon)
{
	Json::Value v(Json::arrayValue);
	OsmGpsMapPoint* p1 =  osm_gps_map_point_new_degrees (lat, lon);
	osm_gps_map_track_add_point(m_polytrack, p1);
	v[0]=lon;
	v[1]=lat;

	m_area[m_area.size()]= v;
}

void MapViewer::set_place(float lat, float lon)
{
	osm_gps_map_gps_clear (m_map);
	osm_gps_map_gps_add (m_map, lat, lon, gtk_spin_button_get_value (m_angle));
	m_place = Json::Value(Json::arrayValue);
	m_place[0]=lon;
	m_place[1]=lat;
}

void MapViewer::new_poly()
{
	m_area = Json::Value(Json::arrayValue);
	if(m_polytrack)
	{
		osm_gps_map_track_remove(m_map, m_polytrack);
		g_object_unref (G_OBJECT(m_polytrack));
	}
	if(m_poly)
	{
		g_object_set(m_poly, "track", NULL, NULL);
		g_object_unref (G_OBJECT(m_poly));
	}
	m_poly=osm_gps_map_polygon_new();
	m_polytrack = osm_gps_map_track_new();

	g_object_set(m_poly, "track", m_polytrack, NULL);
	g_object_set(m_poly, "editable", TRUE, NULL);
	osm_gps_map_polygon_add(m_map, m_poly);
	if(m_init) m_zone->on_save();
}

void MapViewer::new_place()
{
	m_place = Json::Value(Json::arrayValue);
	osm_gps_map_gps_clear (m_map);
	if(m_init) m_zone->on_save();
}

void MapViewer::setAngle(float a)
{
	setUpPlace(m_place);
}

void MapViewer::setUpArea(Json::Value& v)
{
	//std::cout << "Set up area" << v <<std::endl;
	new_poly ();
	for(int i=0; i<v.size(); i++)
	{
		add_point_to_poly((float)v[i][1].asDouble(),(float)v[i][0].asDouble() );
	}
	if(m_init) m_zone->on_save();
}

void MapViewer::setUpPlace(Json::Value& v)
{
	//std::cout << "Set up place:" << v << std::endl;
	printf("Place: %d : %f %f\n", v.size(),(float)v[1].asDouble(),(float)v[0].asDouble());
	set_place((float)v[1].asDouble(),(float)v[0].asDouble());
	if(m_init) m_zone->on_save();
}


Json::Value MapViewer::getArea()
{
	GSList* l = osm_gps_map_track_get_points (m_polytrack);
	Json::Value v(Json::arrayValue);
	while(l)
	{
		float lat, lon;
		Json::Value x(Json::arrayValue);
		OsmGpsMapPoint* p = (OsmGpsMapPoint*) l->data;
		osm_gps_map_point_get_degrees(p, &lat, &lon);
		x[0]=lon;
		x[1]=lat;
		v[v.size()]=x;
		l=l->next;
	}
	return v;
}